@@ -23,7 +23,7 @@ class GithubFileSystem(AbstractFileSystem):
       may specify sha in the extra args
     - 'github://org:repo@/precip/catalog.yml', where the org and repo are
       part of the URI
-    - 'github://org:repo@sha/precip/catalog.yml', where tha sha is also included
+    - 'github://org:repo@sha/precip/catalog.yml', where the sha is also included
 
     ``sha`` can be the full or abbreviated hex of the commit you want to fetch
     from, or a branch or tag name (so long as it doesn't contain special characters
@@ -156,15 +156,17 @@ class GithubFileSystem(AbstractFileSystem):
             if r.status_code == 404:
                 raise FileNotFoundError(path)
             r.raise_for_status()
+            types = {"blob": "file", "tree": "directory"}
             out = [
                 {
                     "name": path + "/" + f["path"] if path else f["path"],
                     "mode": f["mode"],
-                    "type": {"blob": "file", "tree": "directory"}[f["type"]],
+                    "type": types[f["type"]],
                     "size": f.get("size", 0),
                     "sha": f["sha"],
                 }
                 for f in r.json()["tree"]
+                if f["type"] in types
             ]
             if sha in [self.root, None]:
                 self.dircache[path] = out
